/*
 * Copyright (C) 1998-2002 Wolfgang Moser <womo_(at)_d81_(dot)_de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Commodore CBM 1581 floppy disk copy util for PC's, FRMHEADR.C
 */

#include "frmheadr.h"

// Run length encoded data blocks of different empty images
//
// Run lengths of up to 2 bytes are not encoded, run lengths of
// up to 257, 65537, 2^24+1 and 2^32+1 bytes are encoded with
// 4 different tokens. If the input data contains every byte
// representation, another token is defined to mask out these
// values.
//
// The encoding scheme is as following:
//   - calculate the biggest run lenght for every byte value
//   - calculate how many tokens are needed to representate
//     this length, e.g.:
//            biggest length =    234 -->  1 token needed
//               "       "   = 138945 -->  3 tokens needed
//	  - calculate a frequency distribution of all bytes of the
//     input data
//   - get the least used bytes out of the frequency distribution,
//     you have to get out as many values as tokens are needed
//   - if the frequency of one of these values is _not_ null, the
//     next least used byte has also to be got; this gets a token
//     for masking out the token values
//
//   - write the descriptor byte, the lower 3 bits contain the
//     number of run length tokens used, the 5th bit (0x10) tells
//     the decompressor, if another token is used for masking out
//     any of the tokens defined
//   - write the tokens in big endian order (MSB first, LSB last),
//     you can define yourself the values for the desired tokens,
//     but take aware, that you write the token first, that
//     represents the biggest run length; the token representing
//     a maximum run length of 257 is always the last one, if not
//     the only token needed
//   - last write the token for masking out other tokens, if such
//     one is needed
//
//   - compress the data, if the next single or the next two input
//     bytes (up to two byte aren't encoded) have a value of one
//     of the defined tokens, you have to precede the
//     "mask out token" before _every_ of these bytes
//   - if you detect a run of more than 2 bytes in the input data
//     determine the exact length and substract this value by two,
//     next determine the token, that is needed to encode this
//     length and write it to the output stream, write the length
//     (exact length minus two) of this run into the output
//     stream, encode the value in big endian order; don't write
//     preceeding null bytes, you must only use as much bytes as
//     the choosed token allows. Last write the value of the byte
//     to be encoded into the output stream, if this value is the
//     same as one of the tokens, _no_ extra handling is needed,
//     you mustn't preceede this byte with the "mask out token"
//
// The rest would probably become clear, if some plays around with
// the decoder routine, it's really a very simple encoding scheme.
// But if it's to hard, why not imlementing a new and much better
// (de)-compressor for 1581-Copy?
//

// This represents the first part of the BAM of an empty D81 disk image
//
const unsigned char D81_BAM[]={
     2,  4,  1, 40,  3, 68,  0, 49, 53, 56, 49, 32, 67, 79, 80, 89,
    32, 70, 79, 82, 77, 65, 84,160,160, 48, 48,160, 51, 68,160,160,
     1,225,  0, 40,  2, 68,187, 48, 48,192,  1,  7,  0,  1,232,255,
    36,240,  1,  2,255,  0,255, 68,187, 48, 48,192,  1,  7,  0,  1,
   232,255, 40,  1,  3,255,  0,255,  4, 16,252,  0 };

// This represents the BAM of an empty D2M disk image
//
const unsigned char D2M_BAM[]={
     2,  3,  2,  2,254,  0,  1, 34, 72,  0, 49, 53, 56, 49, 32, 67,
    79, 80, 89, 32, 70, 79, 82, 77, 65, 84,160,160, 48, 48,160, 49,
    72,160,160,  2,  1,  0,  1,  1,  2,222,  0, 72,183, 48, 48,192,
     0, 25,  2, 25,  0, 31,  3, 31,217,255,  0,255,  3,  5,252,  0
    };

// This represents the system partition of an empty D2M disk image
//
const unsigned char D2M_SYS[]={
     2,  5,  4,  5,  4,255,  0,  4, 53,255,  0,  0,  4, 52,255,  0,
    12,  4, 52,255,  0,128,  4, 52,255,  0,  0,  1,  1,  4, 10,  0,
    67, 77, 68, 32, 70, 68, 32, 83, 69, 82, 73, 69, 83,  4,  1, 32,
     5,  1,254,  0,  1,  1,255,  0,  0, 83, 89, 83, 84, 69, 77,  4,
     8,160,  4, 11,  0,  1,  0,  0, 80, 65, 82, 84, 73, 84, 73, 79,
    78, 32, 49,  4,  3,160,  4,  7,  0, 12,128,  4,190,  0,  1,  2,
     4,252,  0,  1,  3,  4,253,  0,255,  5, 28,252,  0 };

// decompress an input buffer into an output buffer
//
// input:  in:	      a pointer to the buffer location, that contains
//							the compressed input data
//         size:     the size of the input buffer
//
// output: out       a pointer to the buffer location, where the
//                   decompressed data is stored
//
void deRLE(const unsigned char *in, int size, unsigned char huge *out){
	int i, jump[5];
   unsigned char signs[5], maxS, c;
   unsigned long int len;
   enum jmpConsts{
	   Wrong=-1,
      Subst=5,
      Pow1=1,
      Pow2=2,
      Pow3=3,
      Pow4=4
	   };

	c=*in++;
	size--;
	for(i=4;i>=0;--i) jump[i]=Wrong;

   if(c){
	   maxS=(c&0x07)+((c&0x10)?1:0);
      for(i=0;i<maxS;++i) signs[i]=*in++, size--;
		for(;i<5;i++) signs[i]=*signs;

		if(c&0x10) jump[maxS-1]=Subst;
      for(i=maxS-((c&0x10)?2:1);i>=0;--i) jump[i]=maxS-((c&0x10)?1:0)-i;
      }

   for(;size>0;size--){
   	c=*in++;
   		// if c is in signs
   	for(i=maxS-1;i>=0;--i) if(signs[i]==c) break;
      if(i>=0){	// c is in signs
      	switch(jump[i]){
				default:
				case Wrong:
               return;
				case Subst:
            	*out++=*in++, size--;
               break;
				case Pow4:
				case Pow3:
				case Pow2:
				case Pow1:
					for(len=0ul,i=jump[i];i>0;--i) len<<=8, len|=*in++, size--;
					c=*in++, size--;
					for(len+=2;len;--len) *out++=c;
         	}
      	}
      else *out++=c;
   	}
	}

// create an empty CMD FD2000 style BAM
//
// input:  buffer:	a pointer to the buffer location,
//							where the BAM should be created (0x2400 bytes)
// output: none
//
void FD2000createBAM(unsigned char huge *buffer){
	deRLE(D2M_BAM,	sizeof(D2M_BAM), buffer);
	}

// create a CMD FD2000 style system partition
//
// input:  buffer:	a pointer to the buffer location,
//							where the partition should be created (0x0c00 bytes)
// output: none
//
void FD2000createSYS(unsigned char huge *buffer){
	deRLE(D2M_SYS, sizeof(D2M_SYS), buffer);
	}

// create an empty CBM 1581 style BAM
//
// input:  buffer:	a pointer to the buffer location,
//							where the BAM should be created (0x400 bytes)
// output: none
//
void C1581createBAM(unsigned char huge *buffer){
	int i;
   unsigned char huge *p1;

	deRLE(D81_BAM, sizeof(D81_BAM), buffer);

   p1=buffer+0x110;
   for(i=0x26;i>=0;--i){
		*p1=*(p1+0x100)=0x28;
      p1+=6;
      }
	}

