/**
 * @file c2n.c
 * The main program
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/* Copyright  2001 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#undef HAS_SERIAL
#if defined __unix || defined __unix__
# include <termios.h>
# include <unistd.h>
# include <fcntl.h>
# define HAS_SERIAL
#endif /* defined __unix || defined __unix__ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#include <math.h>

#include "c2n.h"
#include "decode.h"
#include "encode.h"

/** Flag: verbose mode */
unsigned verbose = 0;

/** Name of the current file */
static const char* currentFile;
/** Maximum number of encountered errors */
static unsigned error_thresold = 10;
/** Number of encountered errors */
static unsigned error_count;

/** Error reporter
 * @param pe	the unexpected pulse or error code
 * @param block	the block in which the error occurred (0=first header)
 * @param pos	number of bytes decoded from the stream
 * @return	nonzero on fatal error
 */
static unsigned
error_report (unsigned pe, unsigned block, unsigned pos)
{
  if (currentFile)
    fputs (currentFile, stderr), fputs (": ", stderr);
  switch (pe) {
  case Pause:
    fputs ("unexpected break in the pulse stream", stderr);
    break;
  case Short:
    fputs ("unexpected short pulse", stderr);
    break;
  case Medium:
    fputs ("unexpected medium pulse", stderr);
    break;
  case Long:
    fputs ("unexpected long pulse", stderr);
    break;
  case Parity:
    fputs ("parity error", stderr);
    break;
  case Checksum:
    fputs ("block checksum mismatch", stderr);
    break;
  case NoFirst:
    fputs ("first copy is missing", stderr);
    break;
  case NoSecond:
    fputs ("second copy is missing", stderr);
    break;
  case LongBlock:
    fprintf (stderr, "block %u is %u bytes longer than expected\n",
	     block, pos);
    return 1;
  case ShortBlock:
    fprintf (stderr, "block %u is %u bytes shorter than expected\n",
	     block, pos);
    return 1;
  case Unexpected:
    fprintf (stderr, "unexpected header tag 0x%02x in block %u\n",
	     pos, block);
    goto thresold;
  case Mismatch:
    fputs ("mismatch between first and second copy", stderr);
    break;
  case Countdown:
    fprintf (stderr, "unrecognized countdown byte 0x%02x at block %u\n",
	     pos, block);
    return 1;
  default:
    fputs ("unexpected internal error", stderr);
    break;
  }
  fprintf (stderr, " at %u decoded bytes of block %u\n", pos, block);

 thresold:
  if (++error_count > error_thresold) {
    fprintf (stderr, "more than %u errors; aborted\n", error_thresold);
    return 1;
  }
  return 0;
}

#ifdef HAS_SERIAL
/** file descriptor for the serial device (0=none) */
static int serialfd = 0;
/** previous terminal input/output settings for the serial line */
static struct termios tio;
/** new terminal input/output settings for the serial line */
static struct termios newtio;
#endif /* HAS_SERIAL */

/** A PCM sample of a pulse */
struct pcm_pulse
{
  char* data;		/**< the data bytes */
  unsigned length;	/**< length of the data bytes */
};

/** PCM sample types */
enum pcm_type
{
  eight,		/**< eight bits */
  lsb16,		/**< sixteen bits, least significant byte first */
  msb16,		/**< sixteen bits, most significant byte first */
  eight_,		/**< eight bits, inverted phase */
  lsb16_,		/**< sixteen bits, LSB first, inverted phase */
  msb16_		/**< sixteen bits, MSB first, inverted phase */
};

/** PCM samples of the pulses */
static struct pcm_pulse pcm_short, pcm_medium, pcm_long;
/** File descriptor for PCM output */
static FILE* pcm_file;

/** Compute a stream of pulse code modulation (PCM) samples for a pulse
 * @param sample	the sample stream to be computed
 * @param pulserate	the sample rate (in Hz) multiplied
 *			by the pulse width (in 8-microsecond units)
 * @param bits		number of bits to compute
 * @param sign		flag: unsigned or signed pulse stream
 * @param channels	number of output channels
 * @return		the computed sample
 */
static struct pcm_pulse*
pcm_compute (struct pcm_pulse* sample,
	     unsigned pulserate,
	     enum pcm_type bits,
	     int sign,
	     unsigned channels)
{
  double arg, inc;
  char* c;
  short mult = 0, bias;
  static const double twopi = 2 * 3.14159265358979323846;

  sample->length = pulserate / 125000 * channels;

  switch (bits) {
  case eight:
    mult = 127; break;
  case eight_:
    mult = -127; break;
  case lsb16: case msb16:
    sample->length *= 2; mult = 32767; break;
  case lsb16_: case msb16_:
    sample->length *= 2; mult = -32767; break;
  }

  if (sign) {
    switch (bits) {
    case eight:
    case eight_:
      bias = -128; break;
    default:
      bias = -32768; break;
    }
  }
  else
    bias = 0;

  free (sample->data);
  if (!(sample->data = malloc (sample->length)))
    return 0;

  for (arg = 0, inc = twopi / (pulserate / 125000), c = sample->data;
       c < sample->data + sample->length; arg += inc) {
    short s = bias + mult * sin (arg);
    unsigned ch = channels;
    switch (bits) {
    case eight: case eight_:
      while (ch--) *c++ = s;
      break;
    case lsb16: case lsb16_:
      while (ch--) { *c++ = s; *c++ = s >> 8; }
      break;
    case msb16: case msb16_:
      while (ch--) { *c++ = s >> 8; *c++ = s; }
      break;
    }
  }
  return sample;
}

/** Cleanup function */
static void
cleanup (void)
{
#ifdef HAS_SERIAL
  if (serialfd > 0) {
    tcsetattr (serialfd, TCSADRAIN, &tio);
    tcsendbreak (serialfd, 0);
    close (serialfd);
    serialfd = 0;
  }
#endif /* HAS_SERIAL */

  if (pcm_file)
    fclose (pcm_file);
  free (pcm_short.data), free (pcm_medium.data), free (pcm_long.data);
}

/** Signal handler
 * @param num	the signal number
 */
static void
sig (int num)
{
  cleanup ();
  signal (num, SIG_DFL);
  raise (num);
}

#ifdef HAS_SERIAL
/** Read a pulse from the serial line
 * @return	the read pulse
 */
static enum pulse
sp_read (void)
{
  char c;
  if (!serialfd)
    return Pause;
 again:
  if (!read (serialfd, &c, 1)) {
    cleanup ();
    return Pause;
  }
  switch (c) {
  case 'A':
    return Pause;
  case 'B':
    return Short;
  case 'C':
    return Medium;
  case 'D':
    return Long;
  }
  fprintf (stderr, "unrecognized character 0x%02x in pulse input stream\n", c);
  goto again;
}
#endif /* HAS_SERIAL */

/** Read a pulse from standard input
 * @return	the read pulse
 */
static enum pulse
p_read (void)
{
  int c = getchar ();
  switch (c) {
  case EOF:
  case 'A':
    return Pause;
  case 'B':
    return Short;
  case 'C':
    return Medium;
  case 'D':
    return Long;
  }
  fprintf (stderr, "unrecognized character 0x%02x in pulse input stream\n", c);
  return Pause;
}

/** pulse look-up table */
static const char pulses[] = { 'A', 'B', 'C', 'D' };

/** Write a pulse to the serial line or to standard output
 * @param p	the pulse to be written
 */
static void
p_write (enum pulse p)
{
  if (p >= sizeof pulses)
    return;
#ifdef HAS_SERIAL
  if (serialfd) {
    /** output buffer */
    static char buf[8192];
    /** number of bytes in the output buffer */
    static unsigned datalen;

    buf[datalen++] = pulses[p];
    if (p == Pause || datalen == sizeof buf) {
      write (serialfd, buf, datalen);
      datalen = 0;
    }
  }
  else
#endif /* HAS_SERIAL */
    putchar (pulses[p]);
}

/** Write audio samples corresponding to a pulse to standard output
 * @param p	the pulse to be written
 */
static void
p_write_pcm (enum pulse p)
{
  const struct pcm_pulse* pl;
  switch (p) {
  default:
    return;
  case Short:
    pl = &pcm_short;
    break;
  case Medium:
    pl = &pcm_medium;
    break;
  case Long:
    pl = &pcm_long;
    break;
  }

  fwrite (pl->data, pl->length, 1, pcm_file ? pcm_file : stdout);
}

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful
 */
int
main (int argc, char** argv)
{
  /** mode of operation */
  enum { Encode, Decode } mode = Encode;
  /** flag: raw operation (input and output pulse streams) */
  int raw = 0;
  /** program name */
  const char* prog = *argv;
  /** short pulse width (unit=8 microseconds) */
  static unsigned pulse_short = 0;
  /** medium pulse width (unit=8 microseconds) */
  static unsigned pulse_medium = 0;
  /** long pulse width (unit=8 microseconds) */
  static unsigned pulse_long = 0;
  /** initial sync pulse count */
  static unsigned sync_begin = 1500; /* over 26000 in the KERNAL routines */
  /** intra-block sync pulse count */
  static unsigned sync_intra = 0;
  /** sample rate in Hz */
  static unsigned samplerate = 0;
  /** flag: use plus/4 format */
  static unsigned plus4 = 0;

  atexit (cleanup);
  signal (SIGINT, sig);

  /* process the options */
  for (argv++; argc > 1 && **argv == '-'; argv++, argc--) {
    const char* opts = *argv;

    if (!strcmp (opts, "--")) { /* disable processing further options */
      argv++;
      argc--;
      break;
    }

    while (*++opts) /* process all flags */
      switch (*opts) {
#ifdef HAS_SERIAL
      case 'c':
	if (argc <= 2 || serialfd)
	  goto Usage;
	if ((serialfd = open (*++argv, O_RDWR | O_NOCTTY)) < 0) {
	  fputs ("-c ", stderr);
	  fputs (*argv, stderr);
	  perror (": open");
	  return 2;
	}

	if (tcgetattr (serialfd, &tio)) {
	  fputs ("-c ", stderr);
	  fputs (*argv, stderr);
	  perror (": tcgetattr");
	  close (serialfd);
	  serialfd = -1;
	  return 2;
	}
	else {
	  memcpy (&newtio, &tio, sizeof newtio);
	  newtio.c_iflag = IGNBRK | IGNPAR | IXON;
	  newtio.c_oflag = 0;
	  newtio.c_cflag = B38400 | CS8 | CLOCAL | CREAD;
	  newtio.c_lflag = 0;
	  memset (newtio.c_cc, 0, sizeof newtio.c_cc);
	  newtio.c_cc[VSTART] = 021; /* DC1, ctrl-q */
	  newtio.c_cc[VSTOP] = 023; /* DC3, ctrl-s */
	  newtio.c_cc[VMIN] = 10; /* read at least 10 chars */
	  newtio.c_cc[VTIME] = 1; /* 1/10th second timeout */
	  tcflush (serialfd, TCIOFLUSH);
	  tcsendbreak (serialfd, 0);
	  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
	    fputs ("-c ", stderr);
	    fputs (*argv, stderr);
	    perror (": tcsetattr");
	  c2n232fail:
	    close (serialfd);
	    serialfd = -1;
	    return 2;
	  }
	  else {
	    char response[10];
	    int len;
	    /* send a NUL character and await a response */
	    write (serialfd, "", 1);
	    len = read (serialfd, response, sizeof response);
	    if (!len) {
	      fputs ("-c ", stderr);
	      fputs (*argv, stderr);
	      if (errno)
		perror (": read");
	      else
		fputs (": no response from C2N232 device\n", stderr);
	      goto c2n232fail;
	    }
	    if (response[len - 1] != '0') {
	      fputs ("-c ", stderr);
	      fputs (*argv, stderr);
	      fputs (": unexpected response from C2N232 device\n", stderr);
	      goto c2n232fail;
	    }

	    newtio.c_cc[VMIN] = 0; /* read at least 0 characters */
	    newtio.c_cc[VTIME] = 20; /* wait 2 seconds for end of stream */
	    tcsetattr (serialfd, TCSANOW, &newtio);
	  }
	}
	argc--;
	break;
#endif /* HAS_SERIAL */
      case 'd':
	mode = Decode;
	break;
      case 'e':
	mode = Encode;
	break;
      case 'E':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  /** number of output channels */
	  unsigned channels = 1;
	  /** flag: signed output (as opposed to unsigned) */
	  int sign = 1;
	  /** the sample rate */
	  int rate = strtol (*++argv, &r, 0);
	  enum pcm_type t = rate > 0 ? eight : eight_;
	  if (pcm_file)
	    fclose (pcm_file);
	  pcm_file = 0;
	  argc--;
	  if (!rate) {
	  EUsage:
	    fprintf (stderr, "-E: [+-]rate[blsu]*[,file] expected, got %s\n",
		     *argv);
	    goto Usage;
	  }
	Enext:
	  switch (*r++) {
	  case 0:
	    break;
	  case ',':
	    if (!*r)
	      goto EUsage;
	    if (!(pcm_file = fopen (r, "wb"))) {
	      fputs ("-E ...,", stderr);
	      fputs (r, stderr);
	      perror (": fopen");
	      return 2;
	    }
	    break;
	  case 'b':
	    t = rate > 0 ? msb16 : msb16_;
	    goto Enext;
	  case 'l':
	    t = rate > 0 ? lsb16 : lsb16_;
	    goto Enext;
	  case 's':
	    channels = 2;
	    goto Enext;
	  case 'u':
	    sign = 0;
	    goto Enext;
	  default:
	    goto EUsage;
	  }

	  samplerate = rate < 0 ? -rate : rate;

	  pcm_compute (&pcm_short, samplerate *
		       (pulse_short ? pulse_short : (plus4 ? 60 : 45)),
		       t, sign, channels);
	  pcm_compute (&pcm_medium, samplerate *
		       (pulse_medium ? pulse_medium : (plus4 ? 120 : 63)),
		       t, sign, channels);
	  pcm_compute (&pcm_long, samplerate *
		       (pulse_long ? pulse_long : (plus4 ? 240 : 83)),
		       t, sign, channels);

	  mode = Encode;
	}
	break;
      case 'h':
	goto Usage;
      case '2':
	plus4 = 1;
	break;
      case 'p':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  char* s = *++argv;
	  argc--;
	  pulse_short = strtoul (s, &r, 0);
	  if (*r != ',') {
	  pUsage:
	    fprintf (stderr, "-p: short,medium,long expected, got %s\n",
		     *argv);
	    goto Usage;
	  }
	  s = r + 1;
	  pulse_medium = strtoul (s, &r, 0);
	  if (*r != ',')
	    goto pUsage;
	  s = r + 1;
	  pulse_long = strtoul (s, &r, 0);
	  if (*r)
	    goto Usage;
	}
	break;
      case 'r':
	raw = 1;
	break;
      case 's':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  char* s = *++argv;
	  argc--;
	  sync_begin = strtoul (s, &r, 0);
	  if (*r != ',') {
	  sUsage:
	    fprintf (stderr, "-s: begin,intra expected, got %s\n",
		     *argv);
	    goto Usage;
	  }
	  s = r + 1;
	  sync_intra = strtoul (s, &r, 0);
	  if (*r)
	    goto sUsage;
	}
	break;
      case 't':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* s;
	  error_thresold = strtoul (*++argv, &s, 0);
	  argc--;
	  if (*s) {
	    fprintf (stderr, "-t: non-numeric argument %s\n", *argv);
	    goto Usage;
	  }
	}
	break;
      case 'v':
	verbose++;
	break;

      default:
	goto Usage;
      }
  }

  if (argc < 2) {
  Usage:
    fprintf (stderr,
	     "C2N 1.0.1 - Commodore archive converter\n"
	     "Usage: %s [options] file(s)\n", prog);
    fputs ("Options: -2: use Commodore 264 series pulse widths and format\n"
#ifdef HAS_SERIAL
	   "         -c device: specify the name of the C2N232 device\n"
#endif /* HAS_SERIAL */
	   "         -d: decode pulses to C2N binary stream\n"
	   "         -e: encode C2N binary stream to pulses (default)\n"
	   "         -E [+-]rate[blsu]*[,file]: encode C2N binary stream to PCM\n"
	   "         -h: request this help text\n"
	   "         -p short,medium,long: specify pulse lengths for C2N232\n"
	   "         -r: enable raw operation (input and output pulses)\n"
	   "         -s begin,intra: specify lengths of sync pulse streams\n"
	   "         -t number: specify maximum number of errors\n"
	   "         -v: enable verbose operation\n"
	   "         --: Stop processing any further options.\n",
	   stderr);
    return 1;
  }

#ifdef HAS_SERIAL
  /* Set the pulse widths */
  if (serialfd) {
    unsigned char cmd[5], *op = cmd;
    switch (mode) {
    case Encode:
      *op++ = 3;
      *op++ = 255; /* pause duration */
      *op++ = pulse_short ? pulse_short >> 1 : (plus4 ? 30 : (45 >> 1));
      *op++ = pulse_medium ? pulse_medium >> 1 : (plus4 ? 60 : (63 >> 1));
      *op++ = pulse_long ? pulse_long >> 1 : (plus4 ? 120 : (83 >> 1));
      break;
    case Decode:
      *op++ = 2;
      *op++ = pulse_short ? pulse_short : (plus4 ? 100 : 54);
      *op++ = pulse_medium ? pulse_medium : (plus4 ? 180 : 74);
      *op++ = pulse_long ? pulse_long : (plus4 ? 250 : 94);
      break;
    }

    if (op - cmd != write (serialfd, cmd, op - cmd)) {
      perror ("write");
      return 2;
    }

    if (!read (serialfd, cmd + 1, 1)) {
      if (errno)
	perror ("read");
      else
	fputs ("no response from C2N232 device\n", stderr);
      return 2;
    }

    if ((cmd[1] & ~0x30) != cmd[0]) {
      fprintf (stderr, "C2N232 responded 0x%02x; expected 0x%02x\n",
	       cmd[1], cmd[0] | 0x30);
      return 2;
    }
    if (mode == Encode) {
      newtio.c_cc[VTIME] = 0; /* no waiting at reading */
      if (tcsetattr (serialfd, TCSANOW, &newtio)) {
	perror ("tcsetattr");
	return 2;
      }
    }
  }
#endif /* HAS_SERIAL */

  /* Process the files. */

  for (; --argc; argv++) {
    FILE* df;
    switch (mode) {
    case Encode:
      if (!(df = fopen (currentFile = *argv, "rb"))) {
	fputs (currentFile, stderr);
	perror (": fopen (read)");
	continue;
      }

      if (raw) {
	int c;
	while ((c = fgetc (df)) != EOF) {
	  enum pulse p = Pause;
	  switch (c) {
	  case 'A': break;
	  case 'B': p = Short; break;
	  case 'C': p = Medium; break;
	  case 'D': p = Long; break;
	  default:
	    fprintf (stderr, "%s: unrecognized pulse character 0x%02x\n",
		     currentFile, c);
	    continue;
	  }
	  if (samplerate) p_write_pcm (p); else p_write (p);
	}
      }
      else if (plus4)
	encode264 (df, error_report, samplerate ? p_write_pcm : p_write,
		   sync_begin, sync_intra ? sync_intra : 16384);
      else
	encode (df, error_report, samplerate ? p_write_pcm : p_write,
		sync_begin, sync_intra ? sync_intra : 5376);
      fclose (df);
      break;
    case Decode:
      if (!(df = fopen (currentFile = *argv, "wb"))) {
	fputs (currentFile, stderr);
	perror (": fopen (write)");
	continue;
      }

      error_count = 0;
      if (raw) {
	/* maximum number of successive pauses that will be tolerated */
	static const int maxpauses = 5;
	/* number of pauses that will be tolerated */
	int pauses = maxpauses;
	for (;;) {
	  enum pulse p =
#ifdef HAS_SERIAL
	    serialfd ? sp_read () :
#endif /* HAS_SERIAL */
	    p_read ();
	  if (p == Pause) {
	    if (pauses--)
	      continue;
	    break;
	  }
	  else if (p < sizeof pulses) {
	    if (pauses != maxpauses)
	      fputc (pulses[Pause], df);
	    pauses = maxpauses;
	    fputc (pulses[p], df);
	  }
	}
      }
      else if (plus4)
	fprintf (stderr, "%s: decoded %u bytes (plus/4)\n", currentFile,
		 decode264 (
#ifdef HAS_SERIAL
			    serialfd ? sp_read : p_read,
#else /* HAS_SERIAL */
			    p_read,
#endif /* HAS_SERIAL */
			    error_report, df));
      else
	fprintf (stderr, "%s: decoded %u bytes\n", currentFile,
		 decode (
#ifdef HAS_SERIAL
			 serialfd ? sp_read : p_read,
#else /* HAS_SERIAL */
			 p_read,
#endif /* HAS_SERIAL */
			 error_report, df));
      fclose (df);
      break;
    }
  }

#ifdef HAS_SERIAL
  if (serialfd)
    tcdrain (serialfd);
#endif /* HAS_SERIAL */

  return 0;
}
