/**
 * @file serial.c
 * Serial transfer routines
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  2001 Marko Mkel.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#if defined __unix || defined __unix__
# include <termios.h>
# include <unistd.h>
# include <fcntl.h>
#endif /* defined __unix || defined __unix__ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "info.h"
#include "serial.h"

/** file descriptor for the serial device (0=none) */
static int serialfd = 0;
/** previous terminal input/output settings for the serial line */
static struct termios tio;
/** new terminal input/output settings for the serial line */
static struct termios newtio;
/** Communication buffer */
static unsigned char serial_buf[3];

/** Disable XON/XOFF handshaking
 * @param tios	the terminal input/output structure
 */
static void
disable_xonxoff (struct termios* tios)
{
  tios->c_iflag &= ~IXON;
  tios->c_cc[VSTART] = 0;
  tios->c_cc[VSTOP] = 0;
}

/** Enable XON/XOFF handshaking
 * @param tios	the terminal input/output structure
 */
static void
enable_xonxoff (struct termios* tios)
{
  tios->c_iflag |= IXON;
  tios->c_cc[VSTART] = 021; /* DC1, ctrl-q */
  tios->c_cc[VSTOP] = 023; /* DC3, ctrl-s */
}

/** Receive data with or without calibration
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
static int
serial_do_read (void* buf, unsigned len)
{
  serial_buf[1] = --len;
  serial_buf[2] = len >> 8;
  disable_xonxoff (&newtio);
  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
    perror ("serial_read: tcsetattr");
    return 1;
  }
  if (3 != write (serialfd, serial_buf, 3)) {
    perror ("serial_read: write");
    return 2;
  }
  if (!read (serialfd, serial_buf + 1, 1) ||
      (serial_buf[1] != (serial_buf[0] | '0'))) {
    if (errno)
      perror ("serial_read: read");
    else
      fputs ("serial_read: unexpected response\n", stderr);
    return 3;
  }

  for (len++;;) {
    unsigned got = read (serialfd, buf, len);
    if (!got) {
      if (errno)
	perror ("serial_read: read");
      else
	fputs ("serial_read: truncated data\n", stderr);
      return 4;
    }
    if (got > len) {
      fputs ("serial_read: too much data\n", stderr);
      return 5;
    }
    buf = (char*) buf + got;
    if (!(len -= got))
      break;
  }

  return 0;
}

/** Read computer information
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
static int
serial_detect (struct hostinfo* hostinfo)
{
  unsigned char detectbuf[5];
  if (serial_write ("", 1))
    return 1;
  serial_buf[0] = 6;
  if (serial_do_read (detectbuf, sizeof detectbuf))
    return 2;
  hostinfo->host = detectbuf[0];
  hostinfo->driver = (unsigned) detectbuf[1] | (unsigned) detectbuf[2] << 8;
  hostinfo->basic = (unsigned) detectbuf[3] | (unsigned) detectbuf[4] << 8;
  return 0;
}

/** Open the communication channel
 * @param dev		name of the communication device
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
int
serial_init (const char* dev, struct hostinfo* hostinfo)
{
  if ((serialfd = open (dev, O_RDWR | O_NOCTTY)) < 0) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": open");
    return 1;
  }

  if (tcgetattr (serialfd, &tio)) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": tcgetattr");
    close (serialfd);
    serialfd = -1;
    return 2;
  }

  /* set the communication parameters */
  memcpy (&newtio, &tio, sizeof newtio);
  newtio.c_iflag = IGNBRK | IGNPAR;
  newtio.c_oflag = 0;
  newtio.c_cflag = B38400 | CS8 | CLOCAL | CREAD;
  newtio.c_lflag = 0;
  memset (newtio.c_cc, 0, sizeof newtio.c_cc);
  newtio.c_cc[VMIN] = 10; /* read at least 10 chars */
  newtio.c_cc[VTIME] = 1; /* 1/10th second timeout */

  /* flush the input and output buffers and send a break signal to the line */
  tcflush (serialfd, TCIOFLUSH);
  tcsendbreak (serialfd, 0);

  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
    fputs ("serial_init ", stderr);
    fputs (dev, stderr);
    perror (": tcsetattr");
    close (serialfd);
    serialfd = -1;
    return 2;
  }

  /* send a NUL character and await a response */
  write (serialfd, "", 1);
  do {
    char response[10];
    int len = read (serialfd, response, sizeof response);
    if (!len) {
      fputs ("serial_init ", stderr);
      fputs (dev, stderr);
      if (errno) {
	perror (": read");
      }
      else
	fputs (": no response from C2N232 device\n", stderr);
      continue;
    }
    if (response[len - 1] != '0') {
      fputs ("serial_init ", stderr);
      fputs (dev, stderr);
      fputs (": unexpected response from C2N232 device\n", stderr);
      continue;
    }
    newtio.c_cc[VMIN] = 255; /* read at least 255 characters */
    newtio.c_cc[VTIME] = 10; /* 1-second inter-character timeout */
    tcsetattr (serialfd, TCSANOW, &newtio);
    return serial_detect (hostinfo);
  } while (0);

  serial_close ();
  return 3;
}

/** Close the communication channel */
void
serial_close (void)
{
  if (serialfd > 0) {
    tcsendbreak (serialfd, 0);
    tcsetattr (serialfd, TCSANOW, &tio);
    close (serialfd);
    serialfd = 0;
  }
}

/** Send data
 * @param buf		the data to be sent
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
serial_write (const void* buf, unsigned len)
{
  serial_buf[0] = 4;
  serial_buf[1] = len - 1;
  serial_buf[2] = (len - 1) >> 8;
  enable_xonxoff (&newtio);
  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
    perror ("serial_write: tcsetattr");
    return 1;
  }
  if (3 != write (serialfd, serial_buf, 3)) {
    perror ("serial_write: write(command)");
    return 2;
  }
  if (!read (serialfd, serial_buf, 1) || *serial_buf != '4') {
    if (errno)
      perror ("serial_write: read(command)");
    else
      fputs ("serial_write: unexpected response to command\n", stderr);
    return 3;
  }
  if (len != write (serialfd, buf, len)) {
    perror ("serial_write: write(data)");
    return 4;
  }
  if (!read (serialfd, serial_buf, 1) || *serial_buf != '@') {
    if (errno)
      perror ("serial_write: read(data)");
    else
      fputs ("serial_write: unexpected response to data\n", stderr);
    return 3;
  }
  return 0;
}

/** Receive data
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
serial_read (void* buf, unsigned len)
{
  serial_buf[0] = 5;
  return serial_do_read (buf, len);
}
