
/*
 * This file is a part of the VICE emualtor packages serial.h
 */
/*
 * serial.c - IEEE and serial device implementation.
 *
 * Written by
 *  Teemu Rantanen   (tvr@cs.hut.fi)
 *  Andr Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Patches by
 *  Ettore Perazzoli (ettore@comm2000.it)
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "types.h"
#include "serial.h"
#include "vdrive.h"

static serial_t serialdevices[MAXDEVICES];

/* Store name here for serial-open.  */
#define SERIAL_NAMELENGTH 255

static BYTE SerialBuffer[SERIAL_NAMELENGTH + 1];
static int SerialPtr;

/* On which channel did listen happen to?  */
static int TrapDevice;
static int TrapSecondary;

/* ------------------------------------------------------------------------- */

/* Call this if device is not attached: -128 == device not present.  */
static int fn()
{
    return (0x80);
}

/* These routines work for IEEE488 emulation on both C64 and PET.  */
static int parallelcommand(void)
{
    serial_t *p;
    BYTE b;
    int channel;
    int i, st = 0;

#ifdef DEBUG_SERIAL
    printf("***ParallelCommand %02x %02x %02x  BUF(%d) '%s'\n\n",
         TrapDevice, TrapSecondary, ram[BSOUR], SerialPtr, SerialBuffer);
#endif

    /* which device ? */
    p = &serialdevices[TrapDevice & 0x0f];
    channel = TrapSecondary & 0x0f;

    /* if command on a channel, reset output buffer... */
    if ((TrapSecondary & 0xf0) != 0x60) {
        p->nextok[channel] = 0;
        p->lastok[channel] = 0;
    }
    switch (TrapSecondary & 0xf0) {
      case 0x60:
          /* Open Channel */
          if (!p->isopen[channel]) {
              p->isopen[channel] = 1;
              st = (*(p->openf)) (p->info, NULL, 0, channel);
              for (i = 0; i < SerialPtr; i++)
                  (*(p->putf)) (p->info, SerialBuffer[i], channel);
              SerialPtr = 0;
          }
          if (p->flushf)
              (*(p->flushf)) (p->info, channel);

          if ((!st) && ((TrapDevice & 0xf0) == 0x40)) {
              st = parallelreceivebyte(&b, 1) & 0xbf;   /* any error, except eof
 */
          }
          break;

      case 0xE0:
          /* Close File */
          p->isopen[channel] = 0;
          st = (*(p->closef)) (p->info, channel);
          break;

      case 0xF0:
          /* Open File */
          if (!p->isopen[channel]) {
              p->isopen[channel] = 1;
              SerialBuffer[SerialPtr] = 0;
              st = (*(p->openf)) (p->info, (char *) SerialBuffer, SerialPtr,
                                  channel);
              SerialPtr = 0;

              if (st) {
                  p->isopen[channel] = 0;
                  (*(p->closef)) (p->info, channel);
                  fprintf(stderr, "Serial: Cannot open file. Status $%02x\n",
                          st);
              }
          }
          if (p->flushf)
              (*(p->flushf)) (p->info, channel);
          break;

      default:
          printf("Unknown command %02X\n\n", TrapSecondary & 0xff);
    }
    return (st);
}

int parallelattention(int b)
{
    int st = 0;

    if (pardebug)
        printf("ParallelAttention(%02x)\n", b);

    if (b == 0x3f
        && (((TrapSecondary & 0xf0) == 0xf0)
            || ((TrapSecondary & 0x0f) == 0x0f))) {
        st = parallelcommand();
    } else
        switch (b & 0xf0) {
          case 0x20:
          case 0x40:
              TrapDevice = b;
              break;

          case 0x60:
          case 0xe0:
              TrapSecondary = b;
              st |= parallelcommand();
              break;

          case 0xf0:            /* Open File needs the filename first */
              TrapSecondary = b;
              serialdevices[TrapDevice & 0x0f].isopen[TrapSecondary & 0x0f] = 0;
              if (pardebug)
                  printf("close output.\n");
              break;
        }

    if (!(serialdevices[TrapDevice & 0x0f].inuse))
        st |= 0x80;

    if ((b == 0x3f) || (b == 0x5f)) {
        TrapDevice = 0;
        TrapSecondary = 0;
    }
    if (pardebug)
        printf("ParallelAttention(%02x)->TrapDevice=%02x, st=%04x\n",
               b, TrapDevice, st + (TrapDevice << 8));

    st |= TrapDevice << 8;

    if (attention_callback_func)
        attention_callback_func();

    return st;
}

int parallelsendbyte(int data)
{
    int st = 0;
    serial_t *p;

    p = &serialdevices[TrapDevice & 0x0f];

    if (p->inuse) {
        if (!p->isopen[TrapSecondary & 0x0f]) {

            if (pardebug)
                printf("SerialSendByte[%2d] = %02x\n", SerialPtr, data);
            /* Store name here */
            if (SerialPtr < SERIAL_NAMELENGTH)
                SerialBuffer[SerialPtr++] = data;
        } else {
            /* Send to device */
            st = (*(p->putf)) (p->info, data, TrapSecondary & 0x0f);
        }
    } else {                    /* Not present */
        st = 0x83;
    }

    return st + (TrapDevice << 8);
}

int parallelreceivebyte(BYTE * data, int fake)
{
    int st = 0, secadr = TrapSecondary & 0x0f;
    serial_t *p;

    p = &serialdevices[TrapDevice & 0x0f];

    /* first fill up buffers */
    if (!p->lastok[secadr]) {
        p->lastok[secadr] = p->nextok[secadr];
        p->lastbyte[secadr] = p->nextbyte[secadr];
        p->lastst[secadr] = p->nextst[secadr];
        p->nextok[secadr] = 0;
        if (!p->lastok[secadr]) {
            p->lastst[secadr] =
                (*(p->getf)) (p->info, &(p->lastbyte[secadr]), secadr);
            p->lastok[secadr] = 1;
        }
    }
    if ((!p->nextok[secadr]) && (!p->lastst[secadr])) {
        p->nextst[secadr] =
            (*(p->getf)) (p->info, &(p->nextbyte[secadr]), secadr);
        p->nextok[secadr] = 1;
    }
    *data = p->lastbyte[secadr];
    if (!fake)
        p->lastok[secadr] = 0;
    st = p->nextok[secadr] ? p->nextst[secadr] :
        (p->lastok[secadr] ? p->lastst[secadr] : 2);
    st += TrapDevice << 8;

    if (pardebug)
        printf("receive: sa=%02x lastb = %02x,  ok=%s, st=%04x, nextb = %02x, "
               "ok=%s, st=%04x\n", secadr,
               p->lastbyte[secadr], p->lastok[secadr] ? "ok" : "no",
               p->lastst[secadr],
               p->nextbyte[secadr], p->nextok[secadr] ? "ok" : "no",
               p->nextst[secadr]);

    if ((!fake) && p->nextok[secadr] && p->nextst[secadr])
        p->nextok[secadr] = 0;

    if (pardebug)
        printf("faked = %d, return data =%02x ('%c'), st = %04x\n",
               fake, *data, isprint(*data) ? *data : '_', st);

    if ((st & 0x40) && eof_callback_func != NULL)
        eof_callback_func();
    return st;
}

/* ------------------------------------------------------------------------- */

int serial_init(const trap_t * trap_list)
{
    int i;

#if 0	/* - we don't have any traps :-) */

    /* Remove installed traps, if any.  */
    serial_remove_traps();

    /* Install specified traps.  */
    serial_traps = trap_list;
    serial_install_traps();
#endif
    /*
     * Clear serial device functions
     */
    for (i = 0; i < MAXDEVICES; i++) {
        serial_t *p;
        p = &serialdevices[i];

        p->getf = (int (*)(void *, BYTE *, int)) fn;
        p->putf = (int (*)(void *, BYTE, int)) fn;
        p->openf = (int (*)(void *, char *, int, int)) fn;
        p->closef = (int (*)(void *, int)) fn;
        p->flushf = (void (*)(void *, int)) NULL;
    }

    return 0;
}

int serial_attach_device(int device, char *var, char *name,
                         int (*getf) (void *, BYTE *, int),
                         int (*putf) (void *, BYTE, int),
                         int (*openf) (void *, char *, int, int),
                         int (*closef) (void *, int),
                         void (*flushf) (void *, int))
{
    serial_t *p;
    int i;

    if (device < 0 || device >= MAXDEVICES)
        return 1;

    p = &serialdevices[device];

    if (p->inuse) {
        printf("warning. serial device %d (%s) in use\n", device, name);
        return 1;
    }
    p->getf = getf;
    p->putf = putf;
    p->openf = openf;
    p->closef = closef;
    p->flushf = flushf;

    p->inuse = 1;
    p->info = var;              /* run-time data storage */
    p->name = stralloc(name);

    for (i = 0; i < 16; i++) {
        p->nextok[i] = 0;
    }
    return 0;
}

/* Detach and kill serial devices.  Detach all (shutdown) if device# == -1.  */
int serial_remove(int number)
{
    serial_t *p;
    int i;

    if (number < 0) {
        for (i = 0; i < MAXDEVICES; i++) {

            p = &serialdevices[i];

            if (p && p->inuse) {
                if (number == -2) {     /* QUERY mode */
                    printf("    Unit #%d: %s.\n", i, p->name);  /* add file */
                } else {
#if (defined(DEBUG) || defined(DEBUG_SERIAL))
                    printf("remove unit #%d.\n", i);
#endif
                    serial_remove(i);
                }
            }
        }
    }
    /* number < 0 */
    else {
        if (number < 0 || number >= MAXDEVICES) {
            fprintf(stderr, "\nIllegal device number %d.\n", number);
            return -1;
        }
        p = &serialdevices[number];

        if (!p || !(p->inuse)) {
            printf("I'm not here.  ;)\n");
        } else
            switch (number) {   /* should be based on actual type ... */
#if 0 /*def PRINTER*/
              case 1:
                  tape_detach_image((TAPE *) p->info);
                  break;
              case 4:
              case 5:
                  detach_printout((PRINTER *) p->info);
                  break;
#endif
              case 8:
              case 9:
              case 10:
              case 11:
                  detach_floppy_image((DRIVE *) p->info);
                  break;

              default:
                  printf("Orphan device #%d\n", number);
                  return -1;
            }

    }                           /* else */

    return (0);
}


