/*
 * This file is a part of the VICE emualtor packages serial.h, 
 * and types.h
 */
/*
 * serial.h - Serial device implementation.
 *
 * Written by
 *  Teemu Rantanen  (tvr@cs.hut.fi)
 *  Andr Fachat    (a.fachat@physik.tu-chemnitz.de)
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

#ifndef _SERIAL_H
#define _SERIAL_H

#include <malloc.h>

#include "types.h"

/* ------------------------------------------------------------------------*/

/* dummy replacements, partially from utils.c */

#define	xmalloc(a)	malloc(a)
#define	xrealloc(a,b)	realloc((a),(b))
#define	zopen(a,b,c)	open((a),(b),(c))
#define	zclose(a)	close(a)
#define	is_pc64name(a)	(-1)

#ifndef PATH_MAX
#ifndef POSIX
#define PATH_MAX        MAXPATHLEN
#endif  /* Not POSIX.  */
#endif  /* No PATH_MAX.  */
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif  /* No MAXPATHLEN.  */

/*
 * Binary read and write modes for fopen()
 */

#if (defined(GEMDOS) || defined(__MSDOS__))
#define  READ   "rb"
#define  WRITE  "wb"
#define  APPEND "wb"
#else
#define  READ   "r"
#define  WRITE  "w"
#define  APPEND "w+"
#endif

typedef void trap_t;

extern int pardebug;

static void (*attention_callback_func)(void) = NULL;
static void (*eof_callback_func)(void) = NULL;

static inline char *stralloc(const char *str)
{
    int l = strlen(str);
    char *p = (char *)malloc(l + 1);

    memcpy(p, str, l + 1);
    return p;
}

/* Add the first `src_size' bytes of `src' to the end of `buf', which is a
   malloc'ed block of `max_buf_size' bytes of which only the first `buf_size'
   ones are used.  If the `buf' is not large enough, realloc it.  Return a
   pointer to the new block.  */
static inline char *bufcat(char *buf, int *buf_size, int *max_buf_size,
             const char *src, int src_size)
{
#define BUFCAT_GRANULARITY 0x1000
    if (*buf_size + src_size > *max_buf_size) {
        char *new_buf;

        *max_buf_size = (((*buf_size + src_size) / BUFCAT_GRANULARITY + 1)
                          * BUFCAT_GRANULARITY);
        new_buf = (char *)xrealloc(buf, *max_buf_size);
        buf = new_buf;
    }
    memcpy(buf + *buf_size, src, src_size);
    *buf_size += src_size;
    return buf;
}

/* Set a new value to the dynamically allocated string *str.  */
static inline void string_set(char **str, const char *new_value)
{
    if (*str == NULL) {
        if (new_value != NULL)
            *str = stralloc(new_value);
    } else if (new_value == NULL) {
        free(*str);
        *str = NULL;
    } else {
        *str = xrealloc(*str, strlen(new_value) + 1);
        strcpy(*str, new_value);
    }
}

/* ------------------------------------------------------------------------*/

/* From serial.h */

#define MAXDEVICES              16

/* Serial Error Codes. */

#define SERIAL_OK               0
#define SERIAL_WRITE_TIMEOUT    1
#define SERIAL_READ_TIMEOUT     2
#define SERIAL_FILE_NOT_FOUND   64
#define SERIAL_NO_DEVICE        128

#define SERIAL_ERROR            (2)
#define SERIAL_EOF              (64)


/* Device types. */

#define DT_FS                   0x0100  /* Filesystem Drive */
#define DT_DISK                 0x0200
#define DT_PRINTER              0x0400

#define DT_TAPE                 0x0800

#define DT_MASK                 0xFF


/* Disk Drives. */

#define DT_1540                  0
#define DT_1541                  1      /* Default */
#define DT_1542                  2
#define DT_1551                  3

#define DT_1570                  4
#define DT_1571                  5
#define DT_1572                  6

#define DT_1581                  8

#define DT_2031                 16
#define DT_2040                 17
#define DT_2041                 18
#define DT_3040                 17      /* Same as 2040 */
#define DT_4031                 16      /* Same as 2031 */

/* #define DT_2081                      */

#define DT_4040                 24

#define DT_8050                 32
#define DT_8060                 33
#define DT_8061                 34

#define DT_SFD1001              48
#define DT_8250                 49
#define DT_8280                 50


/* Printers. */

#define DT_ASCII                0       /* No printer commands nor graphics */
#define DT_MPS803               1
#define DT_STAR10CCL            2


typedef struct serial_s
{
    int inuse;
    int isopen[16];                             /* isopen flag for each secondary address */
    char *info;                                 /* pointer to run-time data */
    char *name;                                 /* name of the device */
    int (*getf)(void *, BYTE *, int);           /* serial read function */
    int (*putf)(void *, BYTE, int);             /* serial write function */
    int (*openf)(void *, char *, int, int);     /* serial open function */
    int (*closef)(void *, int);                 /* serial close function */
    void (*flushf)(void *, int);                /* tell device that write completed */
    BYTE nextbyte[16];                          /* next byte to send to emulator, per sec. addr. */
    char nextok[16];                            /* flag if nextbyte is valid */

    int nextst[16];

    /* The PET hardware emulation can be interrupted while
       transferring a byte. Thus we also have to save the byte
       and status last sent, to be able to send it again. */
    BYTE lastbyte[16];
    char lastok[16];
    int lastst[16];
} serial_t;

extern int serial_init(const trap_t *trap_list);

extern int serial_attach_device(int device, char *var, char *name,
                                int (*getf)(void *, BYTE *, int),
                                int (*putf)(void *, BYTE, int),
                                int (*openf)(void *, char *, int, int),
                                int (*closef)(void *, int),
                                void (*flushf)(void *, int));

extern int serial_remove(int number);

extern int parallelattention(int b);
extern int parallelsendbyte(int data);
extern int parallelreceivebyte(BYTE *data, int fake);
extern void serialattention(void);

#endif

