/*******************************************************/
/* rom.c                                               */
/* last change: 28/05/99                               */
/* (c) by Andreas Matthies                             */
/* please report bugs and comments to                  */
/* email: andreas.matthies@arcormail.de                */
/*******************************************************/

#include <string.h>
#include <stdio.h>
#include "rom.h"

static struct rominfo_t rominfo;

/* lsb first */
static const unsigned char rombit[8] = 
	{0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};

/* Header */
static const unsigned char romheaderbyte[9] =
	{137,136,135,134,133,132,131,130,129};

static const unsigned char romprgnamebegin = 5;
static const unsigned char romprgdatabegin = 0;
static unsigned char romheader[72];


static void reset_rom() {
		rominfo.nextheaderbyte = rominfo.lastheaderbyte = 0;
		rominfo.nextbit = 0;
		rominfo.actualbyte = 0;
		rominfo.datachksum = 0;
}


static void init_rom() {
	int byte, bit;
	reset_rom();
	for (byte=0;byte<9;byte++)
		for (bit=0;bit<8;bit++)
			romheader[byte*8+bit] = (romheaderbyte[byte]&rombit[bit]?1:0);
}


static void analyse_rom(prg_t *p,tap_t t) {
	
	char actualbit; 
	char bytefull;
	char couldbeheader;
	long i;
	if ((p->state!=headersearch)&&(p->type!=prg_rom))
		return;

	/* analyse trigger */
	rominfo.halfbit = (rominfo.halfbit?0:1);
	if ((77<=t) && (t<=98)) {
		/* trigger-sync */
		rominfo.halfbit = 0;
	}
	if (!rominfo.halfbit) {
		/* first trigger of the bit; bit is unfinished=4 */
		rominfo.firsttrigger = t;
		actualbit = halfbit;
	}
	else if ((77<=rominfo.firsttrigger) && (rominfo.firsttrigger<=98)) {
		/* byte sync */
		actualbit = bytesync;
	}
	else if ((33<=t) && (t<=76)) {
		/* wide range for the triggervalues */
		/* only comparison of the two triggers important for bit */
		actualbit = ((rominfo.firsttrigger>t)?bit1:bit0);
	}
	else 
		/* no bit of a original tape coding */
		actualbit = wrongtrigger;
    
	/* append bit to actualbyte (only for loading states) */
	bytefull = 0;
	if (actualbit==wrongtrigger) {
		rominfo.nextbit = 0;
		rominfo.actualbyte = 0;
	}
	else if (actualbit==bytesync) {
		rominfo.nextbit = 0;
		rominfo.bitparity = 1;
		
	}
	else if ((actualbit==bit1) || (actualbit==bit0)) {
		if (rominfo.nextbit==8) {
			if (actualbit==bit1)
				rominfo.bitparity = !rominfo.bitparity;
			if (rominfo.bitparity)
				/* bit-parity-error */
				actualbit = parity_error;
			else
				/* bit parity ok */
				actualbit = parity_ok;
			rominfo.nextbit = 0;
			bytefull = 1;
		}
		else if (rominfo.nextbit<=7) {
			rominfo.actualbyte = ((actualbit==bit1)?
				rominfo.actualbyte | rombit[rominfo.nextbit]:
				rominfo.actualbyte & (255-rombit[rominfo.nextbit]));
			rominfo.nextbit++;
			if (actualbit==bit1)
				rominfo.bitparity = !rominfo.bitparity;
		}
	}
	if ((p->state==datasearch) && (p->begin>=p->end)) {
		p->state = headersearch;
		p->changestate = 1;
		printf("error- negative size\n");
	}
	/* searching for header */
	if ((p->state==headersearch)||(p->state==datasearch)) {
		if (actualbit==wrongtrigger) {
			rominfo.nextbit = 0;
			rominfo.nextheaderbyte = rominfo.lastheaderbyte = 0;
		}
		else if ((actualbit==bit0) || (actualbit==bit1)) {
			rominfo.header[rominfo.nextheaderbyte] = actualbit;
			rominfo.nextheaderbyte = (rominfo.nextheaderbyte+1) % 72;
			couldbeheader = (rominfo.nextheaderbyte==rominfo.lastheaderbyte);
			if (couldbeheader) {
				rominfo.lastheaderbyte = (rominfo.lastheaderbyte+1) % 72;
				i = 0;
			}
			/* 72 consequtive bits collected -> check for header */
			while (couldbeheader && (i<72)) {
				couldbeheader = couldbeheader &&
					(romheader[i]==rominfo.header[(rominfo.nextheaderbyte+i)%72]);
					i++;
			}
			if (couldbeheader) {
				/* now it's sure: is is a header!!! */
				p->type = prg_rom;
				p->databegin = romprgdatabegin;
				p->state++;	/* next state */
				p->changestate = 1;
				rominfo.nextbit = 0; /* reset bitcounter */
				rominfo.nextdatabyte = 0;
			}
		}
	}
	else {
		/* loading header/data */
		if (actualbit==wrongtrigger) {
			printf("T");
			p->state = loaderror;
			p->changestate = 1;
		}
		else if (actualbit==parity_error) {
			printf("P");
		}
		if (bytefull) {
			/* byte is full */
			p->data[rominfo.nextdatabyte++] = rominfo.actualbyte;
			if (p->state==headerload) {
				/* check for end of header */
				if (rominfo.nextdatabyte>=romprgnamebegin+16) {
					/* end of header */
					p->begin = p->data[2]*256 + p->data[1];
					p->end = p->data[4]*256 + p->data[3];
					strncpy((char *)&p->name,(char*)&p->data[romprgnamebegin],16);
					p->state = datasearch;
					p->changestate = 1;
				}
			}
			else if ((p->state==dataload) && bytefull) {
				/* check for end of data-block */
				if (rominfo.nextdatabyte > (p->end-p->begin)%0x10000+romprgdatabegin) {
					p->changestate=1;
					rominfo.chksum = p->data[rominfo.nextdatabyte-1];
					if (rominfo.chksum==rominfo.datachksum)
						p->state = loadok;
					else
						p->state = loaderror;
				}
				else if (rominfo.nextdatabyte>romprgdatabegin) {
					/* checksum */
					rominfo.datachksum ^= p->data[rominfo.nextdatabyte-1];
				}
			}
		}
	}
}


struct loader_t rom = {
	reset_rom,
	init_rom,
	analyse_rom
};
