/*******************************************************/
/* tape64.c                                            */
/* (c) by Andreas Matthies                             */
/* please report bugs and comments to                  */
/* email: andreas.matthies@arcormail.de                */
/*******************************************************/

#include "tape64.h"
#include "loader.h"

static param_t parameter;
static unsigned char *inputbuffer;
static unsigned long inputbuffer_last,inputbuffer_next;
static unsigned char *tapbuffer;
static unsigned long tapsinbuffer;
static cbmlist_t *cbmfirst,*cbmlast;


static char *strupr(char *s) {
  char *ret = s;
  while (*s) {
    if ((*s>='a')&&(*s<='z')) {
      *s = *s - 'a' + 'A';
    }
    s++;
  }
  return (s);
}


static BOOL initvars() { 
	/* default parameters */
	parameter.sample_freq = 44100;
	parameter.sample_width = 256; 	/* 8-bit */
	parameter.sample_channels = 1;	/* mono */
	parameter.tap_version = 0;
	parameter.trigamp = 10;
	parameter.tapoutput = 0;		/* no tap-creation */
	parameter.t64output = 0;		/* no t64-creation */
	parameter.p00output = 0;		/* no p00-creation */
	parameter.p00numerate = 0;	/* no p00-numeration */
	parameter.automatic_yes = 0;/* ask for overwrite */
	parameter.reverse_wave = 0;	/* no reverse wave */
	parameter.save_loaderrors = 0;	/* don't save incorrect loaded files */
	parameter.tap_v0_delay_ms = 100; /* 0 in input-tap of version 0 means 100 ms */
	parameter.tapespeed = 1.0;  /* no accelaration */
	cbmfirst = cbmlast = 0;

	/* init buffer for input-file */
	inputbuffer = malloc(sizeof(unsigned char)*MAXINPUTBUFFER);
	inputbuffer_next = inputbuffer_last = 0;
	if (!inputbuffer)
		printf ("couldn't allocate memory\n");
	return ((BOOL)(inputbuffer!=NULL));
}


static int leaveprog(int exitcode) {
	if (inputbuffer)
		free(inputbuffer);
	while (cbmlast = cbmfirst) {
		cbmfirst = cbmfirst->next;
		free(cbmlast);
	}
	return exitcode;
}


static void syntax() {
	printf("tape64 %s (c) %s by Andreas Matthies\n\n",VERSION,LAST_CHANGE);
	printf("Syntax: tape64 -i:<inputfile> [options]\n");
	printf("\n <inputfile> is a sample of a C2N-tape (usually a .wav-file)\n");
	printf("             or a raw tape file (tap-file)\n");
	printf("\n options for sample-input:\n");
	printf("  -f:<frequency> to specify sample-frequency of <inputfile>\n");
	printf("  -t:<trigamp> to specify minimum amplitude of a C2N-trigger\n");
	printf("  -r to reverse the sample in <inputfile>\n");
	printf(" options for tap-input:\n");
	printf("  -v0:<delay> specifies delay in ms for a 0 in tap version 0 (default 100)\n");
	printf(" options for output:\n");
	printf("  -or[:<tap-file>] to create a raw tape file (tap-file)\n");
	printf("  -ot[:<t64-file>] to create a t64-archive\n");
	printf("  -op[:<p00-dir>] to create C64-binaries\n");
	printf("       <p00-dir> must exist and end with \\ (or / in LINUX)\n");
	printf("  -n to numerate the binary-files in tape-order: abc.p00, xyz.p01...\n");
	printf("  -y overwrites files without question\n");
	printf("  -e to save even files with load-errors\n");
	printf("  -s:<speed> to accelerate the tape (default 1.0)");
}


static int getparams(int argc, char *argv[]) {
	int j,i=1;
	char npstr[255];
	int ok = (argc>1);
	if (!ok)
		syntax();
	printf("\n");
	while (ok && (i<argc)) {
		strcpy(npstr,argv[i]);
		strupr(npstr);
		if (npstr[0]=='/')
			npstr[0] = '-';
		/* <inputfile> */
		if (!strncmp(npstr,"-I:",3)) {
			strcpy(parameter.inputname,argv[i]+3);
			strcpy(parameter.name_noextension,argv[i]+3);
			j = strlen(parameter.name_noextension);
			if ((j>=4) && (parameter.name_noextension[j-4]=='.'))
			parameter.name_noextension[j-4] = '\0';
		}
		/* -? */
		else if (!strncmp(npstr,"-?",2)) {
			syntax();
			ok = 0;
		}
		/* -y */
		else if (!strncmp(npstr,"-Y",2))
			parameter.automatic_yes = 1;
		/* -n */
		else if (!strncmp(npstr,"-N",2))
			parameter.p00numerate = 1;
		/* -r */
		else if (!strncmp(npstr,"-R",2))
			parameter.reverse_wave = 1;
		/* -e */
		else if (!strncmp(npstr,"-E",2))
			parameter.save_loaderrors = 1;
		/* -ot[:<t64-filename>] */
		else if (!strncmp(npstr,"-OT",3)) {
			parameter.t64output = 1;
			if (npstr[3]==':')
				strcpy(parameter.t64name,argv[i]+4);
			else
				strcpy(parameter.t64name,"");
		}
		/* -or[:<tap-filename>] */
		else if (!strncmp(npstr,"-OR",3)) {
			parameter.tapoutput = 1;
			if (npstr[3]==':')
				strcpy(parameter.tapname,argv[i]+4);
			else
				strcpy(parameter.tapname,"");
		}
		/* -op[:<p00dir>] */
		else if (!strncmp(npstr,"-OP",2)) {
			parameter.p00output = 1;
			if (npstr[3]==':')
				strcpy(parameter.p00dirname,argv[i]+4);
			else
				strcpy(parameter.p00dirname,"");
		}
		/* -t:trigamp */
		else if (!strncmp(npstr,"-T:",3)) {
			parameter.trigamp = atoi(npstr+3);
			if ((parameter.trigamp<=0) || (parameter.trigamp>=256)) {
				ok = 0;
				printf("-t: wrong value (0<t<256)\n");
			}
		}
		/* -f:<frequency> */
		else if (!strncmp(npstr,"-F:",3)) {
			parameter.sample_freq = atoi(npstr+3);
			if ((parameter.sample_freq<=0) || (parameter.sample_freq>=1000000)) {
				ok = 0;
				printf("-f: wrong value (0 < freq < 1000000)\n");
			}
		}
		/* -s:<speed> */
		else if (!strncmp(npstr,"-S:",3)) {
			parameter.tapespeed = atof(npstr+3);
			if (parameter.tapespeed<=0) {
				ok = 0;
				printf("-s: wrong value (0 < speed)\n");
			}
		}
		/* -v0:<delay> */
		else if (!strncmp(npstr,"-V0:",4)) {
			parameter.tap_v0_delay_ms = atoi(npstr+4);
			if (parameter.tap_v0_delay_ms<=0) {
				ok = 0;
				printf("-v0: wrong value (0 < delay)\n");
			}
		}
		/* wrong parameter */
		else {
			ok = 0;
			printf("wrong parameter: %s\n",argv[i]);
		}
		i++;
	}
	if (parameter.t64output) {
		if (!strlen(parameter.t64name))
			strcpy(parameter.t64name,parameter.name_noextension);
		strcat(parameter.t64name,".t64");
	}
	if (parameter.tapoutput) {
		if (!strlen(parameter.tapname))
			strcpy(parameter.tapname,parameter.name_noextension);
		strcat(parameter.tapname,".tap");
		if (!strcmp(parameter.inputname,parameter.tapname))
			strcat(parameter.tapname,"(copy)");
	}
	return (ok);
}


static check_parameters() {
	int fehler = 0;
	if (parameter.inputtype==sample) {
		/* check sample-parameters */
		if ((parameter.sample_freq<=0) || (parameter.sample_freq>100000)) {
			printf("error: illegal sample-frequence: %d\n",parameter.sample_freq);
			fehler++;
		}
		if ((parameter.sample_width<256) || (parameter.sample_width>256*256)) {
			printf("error: illegal sample-width: %d\n",parameter.sample_width);
			fehler++;
		}
		if ((parameter.sample_channels<1) || (parameter.sample_channels>2)) {
			printf("error: illegal sample-channels: %d\n",parameter.sample_channels);
			fehler++;
		}
	}
	if (parameter.inputtype==sample) {
		if ((parameter.tap_version!=0) && (parameter.tap_version!=1))
			fehler++;
	}
	return (fehler==0);
}


static void post_output() {
	printf("\nResults of processing:\n----------------------\n");
	printf("Input: %s  ",parameter.inputname);
	printf("Type: %s\n",parameter.inputtype==sample?"Sample-File":"TAP-File");
	if (parameter.inputtype==sample) {
		printf("  Triggeramplitude : %d\n",parameter.trigamp);
		printf("  Sample-Frequency : %d\n",parameter.sample_freq);
		printf("  Sample-Range     : 0 - %d\n",parameter.sample_width-1);
		printf("  Channels         : %d\n",parameter.sample_channels);
		printf("  Sample-direction : %s\n",parameter.reverse_wave?"reverse":"normal");
	}
	else if (parameter.inputtype==tap) {
		printf("  Version : %d\n",parameter.tap_version);
		if (parameter.tap_version==0) {
			printf("  Delay for zero-values: %d ms\n",parameter.tap_v0_delay_ms);
		}
	}
	printf("Speed of tape: %3.1f\n",parameter.tapespeed);
	printf("Output: ");
	if (!parameter.tapoutput&&!parameter.t64output&&!parameter.p00output)
		printf("(None)\n");
	else
		printf("\n");
	if (parameter.tapoutput)
		printf("  TAP-File: %s\n",parameter.tapname);
	if (parameter.t64output)
		printf("  T64-File: %s\n",parameter.t64name);
	if (parameter.p00output) {
		printf("  P00's in: %s",parameter.p00dirname[0]?parameter.p00dirname:"Current Dir");
		if (parameter.p00numerate)
			printf(" (numerated)");
		printf("\n");
	}
}



static void pet_to_pc(char *petname) {
	int i,j;
	unsigned char c,cnew,temp[16];
	j=0;
	for (i=0; i<16;i++) {
		c = petname[i];
		cnew = 0;
		if ((65<=c)&&(c<=90))
			cnew = 'a'+c-65;
		else if ((193<=c)&&(c<=218))
			cnew = 'A'+c-193;
		else if (strchr("0123456789!$%&()=',.-_;~+",c))
			cnew = c;
		else if ((c&0x7f) == 32)
			cnew = ' ';
		if (cnew)
			temp[j++] = cnew;
	}
	for (i=0;i<j;i++)
		petname[i] = temp[i];
	petname[j]='\0';
}


static FILE* openfile(char *wname, const char *mode ) {
	char c;
	FILE *f;
	c = 'Y'; 
	if (strchr(mode,'w') && !parameter.automatic_yes && fopen(wname,"r")) {
		printf("%s exists; ok to overwrite? [Y,N] ",wname);
		while (!strchr("YyNn",c = getchar()));
		while (getchar()!='\n');
	}
	if ((c=='Y') || (c=='y')) {
		f = fopen(wname,mode);
		if (!f)
			printf("%s could not be opened!\n",wname);
	}
	else
		f = NULL;
	return f;	
}


static void analyse_inputheader(FILE *wf) {
	tapheader_t tapheader;
	unsigned short header[20],info[100];
	BOOL iswav = 0, istap = 0;
	char wavheader_id[10];
	unsigned long chunklength;
	/* read header */
	if (!fread(header,sizeof(short),6,wf)==6) {
		/* seems to be no good input; we try a sample */
		parameter.inputtype = sample;
		return;
	}
	/* CHECK FOR WAV-HEADER */
	strncpy(wavheader_id,(char *)&header[4],4);
	strupr(wavheader_id);
	if (!strncmp(wavheader_id,"WAVE",4)) {
		/* its a sample-input with wav-header */
		parameter.inputtype = sample;
		do {
			/* read chunk-header */
			if (fread(header,sizeof(short),4,wf)!=4) {
				printf("wav-error: couldn't read chunk-header\n");
				parameter.inputtype = unknown;
			}
			chunklength = *(unsigned long*)&header[2];
			strncpy(wavheader_id,(char *)header,4);
			strupr(wavheader_id);
			if (!strncmp(wavheader_id,"DATA",4))
				return;
			if (!strncmp(wavheader_id,"FMT",3)) {
				/* read fmt-infos */
				if (fread(info,sizeof(char),chunklength,wf)!=chunklength) {
					printf("wav-error: format-info couldn't be read\n");
					parameter.inputtype = unknown;
				}
				if (info[0]!=1) {
					printf("wav-error: seems to be no PCM-format\n");
					parameter.inputtype = unknown;
				}
				parameter.sample_channels = (unsigned char)info[1];
				if (info[7]==8)
					parameter.sample_width = 256;
				else if (info[7]==16)
					parameter.sample_width = 256*256;
				else {
					printf("wav-error: unsupported sample width\n");
					parameter.inputtype = unknown;
				}
				parameter.sample_freq = *(unsigned long*)&info[2];
			}
			else if (fseek(wf,chunklength,SEEK_CUR)) {
				parameter.inputtype = unknown;
			}
		} while (!feof(wf) && (parameter.inputtype==sample));
	}
	rewind(wf);
	/* NOW CHECK FOR TAP-HEADER */
	if (!(fread(&tapheader,sizeof(tapheader_t),1,wf)==1)) {
		/* seems to be no good input; we try a sample */
		parameter.inputtype = sample;
		return;
	}
	if (!strncmp(tapheader.identifier,TAP_IDENTIFIER,12)) {
		/* it is a tap-file */
		parameter.inputtype = tap;
		parameter.tap_version = tapheader.version;
	}
	else {
		/* default: sample */
		rewind(wf);
		parameter.inputtype = sample;
	}
	return;
}


static readinput(unsigned char *i,FILE *wf) {
	if (inputbuffer_next>=inputbuffer_last) {
		if (feof(wf))
			inputbuffer_last = 0;
		else
			inputbuffer_last = fread(inputbuffer,sizeof(unsigned char),MAXINPUTBUFFER,wf);
		inputbuffer_next = 0;
	}
	if (inputbuffer_last)
		*i = inputbuffer[inputbuffer_next++];
	return (inputbuffer_last>0);
}


static void prg_reset(prg_t *p) {
	short l;
	p->state = headersearch;
	p->changestate = 1;
	for (l=0;l<lastloader;l++)
		loaderlist[l]->reset_loader();
	p->begin = p->end = 0;
}


static void prg_message(prg_t *p) {
	char petname[17];
	switch (p->state) {
		case headersearch: 
			printf("searching... ");
			break;
		case headerload:
			printf("found(%d) ",p->type);
			break;
		case datasearch:
			strncpy(petname,p->name,16);
			pet_to_pc(petname);
			printf("%16s at %5d-%5d ",petname,p->begin,p->end);
			break;
		case dataload:
			printf("loading... ");
			break;
		case loadok:
			printf("successful\n");
			break;
		case loaderror:
			printf("load error\n");
			break;
	}
}


static void prg_analyse(prg_t *p,tap_t t) {
	short l;
	for (l=0; l<lastloader; l++)
		loaderlist[l]->analyse_loader(p,t);
}



static BOOL cbmlist_save(prg_t *prg) {
	long i;
	cbmlist_t *cnew;
	cnew = malloc(sizeof(cbmlist_t));
	if (!cnew)
		return 0;
	cnew->begin = prg->begin;
	cnew->end = prg->end;
	strncpy(cnew->c64name,prg->name,16);
	for (i=0;i<prg->end-prg->begin;i++)
		cnew->data[i] = prg->data[prg->databegin+i];
	cnew->next=0;
	if (cbmfirst)
		cbmlast->next = cnew;
	else
		cbmfirst = cnew;
	cbmlast = cnew;
}

		
static init_p00mode() {
	FILE *tempfile;
	char tempfilename[256];
	strcpy(tempfilename,parameter.p00dirname);
	strcat(tempfilename,"__test.tmp");
	if (!(tempfile = fopen(tempfilename,"wb"))) {
		printf("error - couldn't access p00-directory %s\n",parameter.p00dirname);
		return 0;
	}
	else {
		fclose(tempfile);
		remove(tempfilename);
		return 1;
	}
}


static finish_p00mode() {
	FILE *binfile;
	char binfilename[256],extension[10],petfile[17];
	cbmlist_t *cbm;
	unsigned short cbmlength;
	short i,number = 0;
	int errors = 0;
	cbm = cbmfirst;
	while (cbm) {
		cbmlength = cbm->end - cbm->begin;
		sprintf(extension,".p%02d",number);
		strcpy(binfilename,parameter.p00dirname);
		strncpy(petfile,cbm->c64name,16);
		pet_to_pc(petfile);
		for (i=15; i>=0; i--)
			if ((petfile[i]==(char)0x20) || (petfile[i]==(char)0xa0))
				petfile[i] = 0;
			else
				i=0;
		strcat(binfilename,petfile);
		strcat(binfilename,extension);
		if (parameter.p00numerate)
			number++;
		if (!(binfile = openfile(binfilename,"wb")) 
			|| !(fwrite(&cbm->begin,sizeof(short),1,binfile) == 1)
			|| !(fwrite(&cbm->data,sizeof(char),cbmlength,binfile) == cbmlength)) {
			printf("error - couldn't write binary file %s\n",binfilename);
			errors++;
		}
		if (binfile)
			fclose(binfile);
		cbm = cbm->next;
	}
	return (!errors);
}


static FILE *init_tapmode(tapheader_t *tap) {
	FILE *tf;
	strncpy(tap->identifier,TAP_IDENTIFIER,12);
	tap->version = 1; /* new for better emulation of delays */
	tap->filesize = 0;
	tapbuffer = malloc(sizeof(sample_t)*MAXTAPBUFFER);
	if (!tapbuffer)
		return NULL;
	tapsinbuffer = 0;
	tf = openfile(parameter.tapname, "wb");
	if (tf)
		fwrite(tap,sizeof(tapheader_t),1,tf);
	return tf;
}


static BOOL save_tap(tap_t tap,FILE *tf, tapheader_t *th) {
	/* now version 1 of tap is used */
	BOOL ok=1;
	int j,i=0;
	unsigned char c[4];
	ok = 1;
	if (tap/8 <= 255)
		c[i++] = (unsigned char)(tap/8);
	else {
		c[i++] = 0;
		c[i++] = (unsigned char)(tap % 256);
		c[i++] = (unsigned char)((tap>>8) % 256);
		c[i++] = (unsigned char)((tap>>16) % 256);
	}
	th->filesize += i;
	for (j=0; (j<i)&&ok; j++) {
		tapbuffer[tapsinbuffer++] = c[j];
		if (tapsinbuffer>=MAXTAPBUFFER) {
			ok = (fwrite(tapbuffer,sizeof(unsigned char),tapsinbuffer,tf)==tapsinbuffer);
			tapsinbuffer = 0;
		}
	}
	return ok;
}


static BOOL finish_tapmode(FILE *tf,tapheader_t *tap) {
	BOOL ok;
	/* flush buffer and write length-information */
	ok = (fwrite(tapbuffer,sizeof(unsigned char),tapsinbuffer,tf)==tapsinbuffer)
		&& !fseek(tf,16L,SEEK_SET)
		&& (fwrite(&tap->filesize,sizeof(long),1,tf));
	fclose(tf);
	free(tapbuffer);
	return ok;
}


static FILE *init_t64mode(t64header_t *t64) {
	int l,i;
	strncpy(t64->identifier,T64_IDENTIFIER,32);
	t64->major_version = 1;
	t64->minor_version = 0;
	t64->max_entries = 0;
	t64->used_entries = 0;
	for (i=0;i<24;i++)
		t64->tape_name[i] = 0x20;
	i = l = strlen(parameter.name_noextension);
	while (i && !strchr("\\/",parameter.name_noextension[i-1]))
		i--;
	if (l-i>16)
		l = i+16;
	strncpy(t64->tape_name,parameter.name_noextension+i,l-i);
	for (i=0;i<16;i++)
		if ((t64->tape_name[i]>='a') && (t64->tape_name[i]<='z'))
			t64->tape_name[i] = t64->tape_name[i] - 'a' + 'A';
	return openfile(parameter.t64name, "wb");
}


static BOOL finish_t64mode(FILE *t64file,t64header_t *t64,short entries) {
	cbmlist_t *cbm;
	unsigned long current_offset;
	t64dir_t t64dir;
	t64dir.c64stype = 1;
	t64dir.filetype = 0x82;
	t64dir.notused = 0;
	t64dir.evennotused = 0;
	t64->used_entries = entries;
	t64->max_entries = t64->used_entries;
	if (fwrite(t64,sizeof(t64header_t),1,t64file)!=1) {
		printf("error: couldn't write t64-header\n");
		return 0;
	}
	cbm = cbmfirst;
	current_offset = sizeof(t64header_t)+t64->max_entries*sizeof(t64dir_t);
	/* Directory of t64 */
	while (cbm) {
		strncpy(t64dir.name,cbm->c64name,16);
		t64dir.begin = cbm->begin;
		t64dir.end = cbm->end;
		t64dir.offset = current_offset;
		if (fwrite(&t64dir,sizeof(t64dir_t),1,t64file)!=1) {
			printf("error: couldn't write t64-directory\n");
			return 0;
		}
		current_offset += cbm->end-cbm->begin;
		cbm = cbm->next;
	}
	cbm = cbmfirst;
	/* Data of t64 */
	while (cbm) {
		if (fwrite(&cbm->data,sizeof(unsigned char),cbm->end-cbm->begin,t64file)
			!=(unsigned long)cbm->end-cbm->begin) {
			printf("error: couldn't write t64-filedata\n");
			return 0;
		}
		cbm = cbm->next;
	}
	fclose(t64file);
	return 1;
}


static void init_samplestate(samplestate_t *s) {
	s->last_direction = down;
	s->last_sample = ((parameter.sample_width==256)?255:65535);
	s->last_trigger = 0;
	s->sample_count = 0;
}


static void init_prg(prg_t *p) {
	p->loadok = 0;
	p->loaderror = 0;
	prg_reset(p);
}


static read_trigger_from_file(tap_t *tr, FILE *f, samplestate_t *ss) {
	unsigned char t1,t2,t3,t4;
	BOOL ok;
	if ((ok = readinput(&t1,f)) && (!t1) && (parameter.tap_version>=1)) {
		ok = (readinput(&t2,f) && readinput(&t3,f) && readinput(&t4,f));
	}
	*tr = t1*8;
	if (!t1) {
		if (parameter.tap_version==0)
			*tr = (unsigned long)(parameter.tap_v0_delay_ms*c64_pulse/1000); /* delay for a 0 */
		if (parameter.tap_version>=1)
			*tr = t2 + (t3<<8) + (t4<<16);
	}
	return ok;
}


static readsample(unsigned char *sa, FILE *f) {
	BOOL ok;
	char channel;
	unsigned char by[2];
	for (channel=0;channel<parameter.sample_channels;channel++) {
		ok = readinput(&by[channel],f);
		if (parameter.sample_width>256) {
			/* 16-bit-sample -> read most significant byte */
			ok = readinput(&by[channel],f);
			by[channel] = (unsigned char)(by[channel] + 128);
		}
	}
	if (parameter.sample_channels==1)
		*sa = by[0];
	else
		*sa = (by[0]+by[1])/2;
	if (parameter.reverse_wave)
		*sa = 255-*sa;
	return ok;
}


static read_trigger_from_sample(tap_t *tr, FILE *f, samplestate_t *ss) {
	double tempf;
	unsigned char samplebyte;
	BOOL found = 0;
	/* read samples till next trigger */
	while (!found && readsample(&samplebyte,f)) {
		ss->sample_count++;
		ss->sample = samplebyte;
		ss->direction = (ss->last_sample>ss->sample)?down:up;
		if ((ss->last_direction==up) && (ss->direction==down)) {
			ss->last_max = ss->last_sample;
			if ((ss->last_max-ss->last_min)>parameter.trigamp) {
				/* trigger identified */
				if (found = (ss->last_trigger>0)) {
					tempf = (double)(ss->sample_count-ss->last_trigger)
						/parameter.sample_freq*c64_pulse;
					*tr = (tempf>=MAXTAPVALUE)?MAXTAPVALUE:(tap_t)tempf;
				}
				ss->last_trigger = ss->sample_count;
			}
		}
		else if ((ss->last_direction==down) && (ss->direction==up))
			ss->last_min = ss->last_sample;
		ss->last_sample = ss->sample;
		ss->last_direction = ss->direction;
	}
	return found;
}



static BOOL get_trigger(tap_t *tr, FILE *f, samplestate_t *ss) {
	if(parameter.inputtype==sample)
		return read_trigger_from_sample(tr,f,ss);
	if(parameter.inputtype==tap)
		return read_trigger_from_file(tr,f,ss);
}

			
int main(int argc, char *argv[]) {
	FILE *inputfile=NULL, *t64file=NULL, *tapfile=NULL;
	BOOL p00_okay;
	samplestate_t samplestate;
	tapheader_t tapheader;
	t64header_t t64header;
	prg_t prg;
	tap_t trigger;
	unsigned long triggercount = 0;
	short toload,l;
	
	/* Initialization for the loaders */
	for (l=0; l<lastloader; l++)
		loaderlist[l]->init_loader();

	if (!initvars() 
		|| !getparams(argc, argv) 
		|| !(inputfile=openfile(parameter.inputname, "rb")))
		return leaveprog(-1);
	analyse_inputheader(inputfile);
	if (!check_parameters())
		return leaveprog(-2);
	if (parameter.p00output) 
		p00_okay = init_p00mode();
	if (parameter.tapoutput)
		tapfile = init_tapmode(&tapheader);
	if (parameter.t64output)
		t64file = init_t64mode(&t64header);
	init_samplestate(&samplestate);
	init_prg(&prg);
	/* now read all the input */
	while (get_trigger(&trigger,inputfile,&samplestate)) {
		trigger = (tap_t)(trigger * parameter.tapespeed);
		triggercount++;
		if (tapfile) {
			save_tap(trigger,tapfile,&tapheader);
		}
		/* prg translation */
		if (prg.changestate) {
			prg.changestate = 0;
			prg_message(&prg);
		}
		if ((prg.state==loadok)||(prg.state==loaderror)) {
		/* cbm-file is finished */
			if (prg.state==loadok)
				prg.loadok++;
			else
				prg.loaderror++;
			if ((prg.state==loadok) || parameter.save_loaderrors)
				cbmlist_save(&prg);
			prg_reset(&prg);
		}
		prg_analyse(&prg,trigger/8); /* analyse is based on tap-values, so trigger/8 */
	}
	fclose(inputfile);
	printf("\n reached end of tape.\n");
	post_output();
	toload = prg.loadok;
	if (parameter.save_loaderrors)
		toload += prg.loaderror;
	if (parameter.tapoutput && (!tapfile || !finish_tapmode(tapfile,&tapheader)))
		printf("  -> error writing tap-file\n");
	if (parameter.t64output && (!t64file || !finish_t64mode(t64file,&t64header,toload)))
		printf("  -> error writing t64-file\n");
	if (parameter.p00output && (!p00_okay || !finish_p00mode()))
		printf("  -> error writing p00-files\n");
	printf("\nC2N-Triggers : %ld\n",triggercount);
	printf("Programs loadable: %d\nPrograms with load error: %d\n",prg.loadok,prg.loaderror);
	if (prg.state!=headersearch)
		printf("Last program didn't finish before end of input file\n");
	return leaveprog(0);
}

