/*
  taphist.c

  Displays a histogram over a TAP-file

  Assumes that int is at least 32 bits.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
  Reads the header of the TAP-file.
  Length of the data is filled into datalen.
  Returns version of TAP-file, or -1 on error.
 */
int readtapheader (FILE *tapfile, int *datalen)
{
    unsigned char tapheader[20];

    if (fread (tapheader, sizeof(char), 20, tapfile) != 20)
    {
        fprintf (stderr, "Error reading tapheader!\n");
        return -1;
    }

    if (strncmp (tapheader, "C64-TAPE-RAW", 12))
        return -1;

    *datalen = tapheader[16] + (tapheader[17]<<8) + (tapheader[18]<<16) 
      + (tapheader[19]<<24);

    return tapheader[12];
}


/*
  Makes a histogram of the pulses in tapfile.
  Pulses longer than 255*8 cycles are counted as 0.
  Returns an allocated array of the histogram.
 */
int *maketaphist (FILE *tapfile, int datalen, int version)
{
    int *hist;
    int c, c2, c3;
    unsigned int idx;

    hist = calloc(255*8+1, sizeof(int));
    if (!hist)
        return NULL;    

    while (datalen)
    {
        c = getc (tapfile);
        if (c == EOF) 
            return hist;

        if (c == 0 && version > 0)
        {
            c = getc (tapfile);
            c2 = getc (tapfile);
            c3 = getc (tapfile);

            if (c == EOF || c2 == EOF || c3 == EOF)
                return hist;
            idx = c + (c2<<8) + (c3<<16);

            if (idx <= 255*8)
                hist[idx]++;
            else
                hist[0]++;
        }
        else
        {
            hist[c*8]++;
            datalen--;
        }
    }

    return hist;
}


/*
  Prints all nonzero entries in the histogram.
*/
void printhist (int *hist)
{
    int i;

    for (i=1; i<=255*8; i++)
    {
        if (hist[i])
            printf ("%5d (%3d) : %d\n", i, i/8, hist[i]);
    } 

    if (hist[0])
        printf ("Overflow    : %d\n", hist[0]);
}


/*
  Saves the histogram as a runnable matlab script that displays the histogram.
*/
int mkmfile (FILE *mfile, int *hist)
{
    int i;
    
    fprintf (mfile, "clear, clf\n");
    fprintf (mfile, "taphist = [\n");
    
    for (i=0; i<=255*8; i++)
    {
        if (hist[i])
            fprintf (mfile, "\t%d %d\n", i, hist[i]);
    }
    fprintf (mfile, "];\n\n");
    fprintf (mfile, "bar (taphist(:,1), taphist(:,2), 1);\n");
    fprintf (mfile, "axis ([0 256*8 0 1.1*max(taphist(:,2))]);\n");
    fprintf (mfile, "shading flat\n");

    return 0;
}

int main (int argc, char **argv)
{
    FILE *tapfile;
    FILE *mfile;
    int datalen;
    int version;
    int *hist;


    if (argc != 2 && argc != 3)
    {
        printf ("Usage:\n"
                "taphist <filename> [m-file]\n");
        return 0;
    }

    tapfile = fopen (argv[1], "rb");
    if (!tapfile)
    {
        fprintf (stderr, "Cannot open input file.\n");
        return 0;
    }

    version = readtapheader (tapfile, &datalen);
    if (version == -1)
    {
        fprintf (stderr, "Not a TAP-file.\n");
        return 0;
    }

    hist = maketaphist (tapfile, datalen, version);
    if (!hist)
    {
        fprintf (stderr, "Error making histogram.\n");
        return 0;
    }

    fclose (tapfile);

    if (argc == 2)
        printhist(hist);
    else
    {
        mfile = fopen (argv[2], "w");
        if (!mfile)
        {
            fprintf (stderr, "Cannot open m-file.\n");
            return 0;
        }
        mkmfile(mfile, hist);
        fclose(mfile);
    }

    free(hist);
    return 0;
}





