#ifdef WIN32
# include <windows.h>
# include <wincon.h>
#else
# include <signal.h>
# include <time.h>
# include <unistd.h>
#endif
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include "tapir.h"
#include "sound.h"
#include "cmdline.h" /* Replaces options.h */

#ifdef WIN32
/* Ugly fix for a really stupid problem */
# define strncasecmp _strnicmp
#endif /* WIN32 */

/* Globals (for settings that apply to the whole program) */
struct gengetopt_args_info options;
double max_level = 0.0;

/* Sound input/output settings structs */
sound_input in;
sound_output out;

void process_buffer(void *data, uint32 length)
{
    uint32 i;
    int16 *sdata = (int16 *) data;
    static double *ddata = NULL;
    double thresh = 200; /* Not a threshold, per se. Used for hysteresis */
    static int state = 0;

    static double pulselen = 0;
    static double lastdata = 0;

    if(ddata == NULL) {
	ddata = malloc(length * sizeof(double));
    }

    if (!ddata)
	return;

    /* The hysteresis value is dependant on the input sample width */
    thresh = (in.sample_format == EIGHT_BITS) ? 2 : 200;

    filter_buffer(sdata, ddata, length, 1, in.sample_format);

    for (i=0; i<length; i++) {
        if (state == 0) {
	    /* positive pulse */
            if (ddata[i] <= -thresh)
                state = 1;
        } else {
            /* negative pulse */
            if (ddata[i] >= thresh) {
                double deltalen;
		
                /* Do a linear interpolation between the last and
                   current points */
                deltalen = (thresh - lastdata) / (ddata[i] - lastdata);
                pulselen += deltalen;

                write_output_data(pulselen * options.speed_arg,
			       (int)in.samplerate);

                state = 0;
		
                /* This is the time between the threshold crossing and
                   the current sample */
                pulselen = -deltalen;
            }
        }
        pulselen+=1;
	lastdata=ddata[i];
	if(options.level_flag) {
	    /* Check for new max value */
	    if(abs(ddata[i]) > max_level) {
		max_level = abs(ddata[i]);
	    } else {
		/* Allow the max to decay over time */
		max_level /= 1.001;
	    }
	}
    }

    if(options.level_flag) {
	printf("Level: %3.0f%%\n", 100*max_level/SHRT_MAX);
    }
    
#if 0
    /* There's really no need to allocate this again every time... */
    free(ddata);
    ddata = NULL;
#endif
}

#ifdef WIN32
/* Callback for sound input */
int sound_process_input(WAVEHDR *header)
{
    process_buffer(header->lpData, header->dwBytesRecorded/2);
    /* We don't give a rat's ass whether it failed or not... */
    return TRUE;
}

/* Callback for sound output. Can't be used yet, so don't do anything bad */
int sound_process_output(WAVEHDR *header)
{
    return TRUE;
}

#elif defined LINUX

/* Callback for sound input */
int sound_process_input(void* in, uint32 size)
{
    process_buffer(((sound_input*)in)->buffer,
		   size / (((sound_input*)in)->sample_format == EIGHT_BITS ? 1 : 2));

    return TRUE;
}

/* Callback for sound output. Can't be used yet, so don't do anything bad */
int sound_process_output(void* in, uint32 size)
{
    return TRUE;
}

void break_handler(int signum) 
{
    DBG("Brk!\n");

    in.exit_flag = TRUE;
}
#endif /* LINUX */

#ifdef WIN32
void perror_wavein(MMRESULT res)
{
    char temp[300];
    waveInGetErrorText(res, temp, 300);
    puts(temp);
}

void perror_message(uint32 error)
{
    char temp[300];

    FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL, error, 0, temp, 300, NULL);
    puts(temp);
}
#elif defined LINUX
void perror_wavein(uint32 res)
{
    char temp[] = "perror_wavein();  *Not implemented*\n";
    puts(temp);
}

void perror_message(int error)
{
/*    char temp[] = "perror_message(); *Not implemented*\n"; */
    fputs(strerror(error), stderr);
}
#endif

int init_stuff (void)
{

    /* Now it's time to do something useful with all the options */

    if (options.loader_given) {
        if (!strncasecmp(options.loader_arg, "tap", 10))
            settings.loader = TAP;
        else if (!strncasecmp(options.loader_arg, "turbo", 10)
                 || !strncasecmp(options.loader_arg, "turbotape", 10))
            settings.loader = TURBO;
        else if (!strncasecmp(options.loader_arg, "kernal", 10)
                 || !strncasecmp(options.loader_arg, "kernel", 10))
            settings.loader = KERNAL;
        else {
            printf("Loader type %s unknown\n", options.loader_arg);
            return FALSE;
        }
    } else {
        settings.loader = TAP;
    }

    if (settings.loader == TAP) {
        if (!options.outfile_given) {
            printf("Output file name required.\n");
            return FALSE;
        }
    } else {
        if (options.outfile_given)
            printf("Warning: output file name ignored.\n");
    }

    if(options.rate_given) {
	in.samplerate = options.rate_arg;
    } else {
	in.samplerate = 48000;
    }

    if(options.mono_flag) {
	in.channels = MONO;
    } else {
	printf("\nSorry, stereo mode not supported yet\n");
	/* info.channels = STEREO; */
	in.channels = MONO;
    }

    if(options.bits_given) {
	if(options.bits_arg == 8) {
	    in.sample_format = EIGHT_BITS;
	} else if(options.bits_arg == 16) {
	    in.sample_format = SIXTEEN_BITS;
	} else {
	    printf("\n%d bits per sample not supported (yet?)\n",
                   options.bits_arg);
	}
    } else {
	in.sample_format = SIXTEEN_BITS;
    }

#ifdef WIN32
    in.handle = 0;
    in.caps = NULL;
    in.format = NULL;
    {
	int i;
	for(i = 0; i < NUM_BUFFERS; i++ ) in.header[i] = NULL;
    }

#elif defined LINUX
    in.fd = 0;
    in.frag_size = 0;
#endif
    in.callback = sound_process_input;

    /* Set up all things concerning processor clock frequency */
    if(options.clock_given) {
	settings.type = 0;
	settings.video = 0;
	settings.clock_freq = options.clock_arg;
    } else if(options.type_given) {
	if(!strncasecmp(options.type_arg, "c64pal", 10)) {
	    settings.type = C64;
	    settings.video = PAL;
	    settings.clock_freq = C64PAL;
	} else if(!strncasecmp(options.type_arg, "c64ntsc", 10)) {
	    settings.type = C64;
	    settings.video = NTSC;
	    settings.clock_freq = C64NTSC;
	} else if(!strncasecmp(options.type_arg, "vic20pal", 10) ||
		  /* Especially for our german friends :) */
		  !strncasecmp(options.type_arg, "vc20pal", 10) ) {
	    settings.type = VIC20;
	    settings.video = PAL;
	    settings.clock_freq = VIC20PAL;
	} else if(!strncasecmp(options.type_arg, "vic20ntsc", 10) ||
		  !strncasecmp(options.type_arg, "vc20ntsc", 10)) {
	    settings.type = VIC20;
	    settings.video = NTSC;
	    settings.clock_freq = VIC20NTSC;
	} else if(!strncasecmp(options.type_arg, "c16pal", 10)) {
	    settings.type = C16;
	    settings.video = PAL;
	    settings.clock_freq = C16PAL;
	} else if(!strncasecmp(options.type_arg, "c16ntsc", 10)) {
	    settings.type = C16;
	    settings.video = NTSC;
	    settings.clock_freq = C16NTSC;
	} else {
	    printf("Computer type %s unknown\n", options.type_arg);
	    return FALSE;
	}
    } else {
	/* Set to the default */
	settings.type = C64;
	settings.video = PAL;
	settings.clock_freq = C64PAL;
    }

    if(!options.speed_given ||
       (options.speed_arg < 0) ||
       (options.speed_arg > 2)) {

	options.speed_arg = 1;
    }

    if(options.verbose_flag) {
	/* Things that might be nice to know */
	printf("Clock frequency: %dHz\n", settings.clock_freq);
	printf("Speed correction factor: %1.2f\n", options.speed_arg);
	printf("Trying to set input to: ");
	printf("Sample rate %d, ", in.samplerate);
	printf("%s, ", in.channels == STEREO ? "Stereo" : "Mono");
	printf("%d bits per sample\n", in.sample_format == EIGHT_BITS ? 8:16);
    }

    if(options.debug_flag) {
	/* Bail out before setting sound input parameters */
	printf("Debug info is taken care of in another way....\n");
	return FALSE;
    }

    if(options.source_given == TRUE) {
	/* Try to choose an input by name */
	if(sound_choose_input(-1, options.source_arg, &in) != TRUE) {
	    printf("\nCan't choose input: %s\n"
		   "Are you sure the name is correct?\n",
		   options.source_arg);
	    return FALSE;
	}

    } else {
	/* No name given, check for number */
	if(options.number_given == TRUE) {
	    /* Try to choose an input by number */
	    if(sound_choose_input(options.number_arg, NULL, &in) != TRUE) {
		printf("\nCan't choose input #%d\n"
		       "Are you sure the number is correct?\n",
		       options.number_arg);
		return FALSE;
	    }
	} else {
	    /* Choose default input */
	    if(sound_choose_input(-1, NULL, &in) != TRUE) {
		printf("Can't choose default input.\n"
		       "Something is seriously wrong with either your setup or\n"
		       "this program.\n");
		return FALSE;
	    }
	}
    }

    if(sound_set_format(&in, &out) != TRUE)
	return FALSE;
    if(sound_open_input(&in) != TRUE) {
	printf("Can't open input. Probably something wrong with the parameters.\n");
	return FALSE;
    }

    /* Change Ctrl-C handling for windows */
    /* Install a signal handler for Linux */
#ifdef WIN32
    {
	DWORD mode;
	HANDLE hstdin=GetStdHandle(STD_INPUT_HANDLE);
	GetConsoleMode(hstdin, &mode);
	mode &= ~ENABLE_PROCESSED_INPUT;
	SetConsoleMode(hstdin, mode);
    }
#elif defined LINUX
    if(signal(SIGTERM, break_handler) == SIG_ERR ||
       signal(SIGINT, break_handler) == SIG_ERR) {
	perror(__FUNCTION__);
	return FALSE;
    }
#endif

    return TRUE;
}

int close_stuff (void)
{
    sound_stop_input(&in);
    sound_close_input(&in);

    return TRUE;
}

void print_current_time(void)
{
#ifdef WIN32
    char temp[] = "print_current_time(); Not implemented\n";
    SYSTEMTIME systime;

    GetLocalTime (&systime);
    GetTimeFormat (LOCALE_USER_DEFAULT, 0, &systime, NULL, temp, 100);

    puts (temp);
#elif defined LINUX
    time_t current_time;

    current_time = time(NULL);
    puts(ctime(&current_time));
#endif
}

int main (int argc, char *argv[])
{
    /* Parse the command line for options to use */
    if(cmdline_parser(argc, argv, &options) != 0)
	return FALSE;

    if(options.verbose_flag) {
	print_current_time();
    }

    if (!init_stuff())
        return FALSE;

    if (!open_output_file(options.outfile_arg)) {
        return FALSE;
    }

    sound_start_input(&in);

#ifdef WIN32
    /* Wait for return */
    {
        HANDLE hstdin=GetStdHandle(STD_INPUT_HANDLE);
        char buf[100];
        uint32 nread;

        ReadFile(hstdin, buf, 1, &nread, NULL);
    }
#elif defined LINUX
    /* Wait for the signal handler to flip the exit_flag */
    while(in.exit_flag != TRUE) {
	sleep(1);
    }
#endif /* WIN32 */
    close_stuff();

    if(options.verbose_flag) {
	print_current_time();
    }

    if (!close_output_file())
        return FALSE;

    return 0;
}

